/* 
 * This file is part of AndSys 0.4-rc3 distributed under the MIT license.
 * 
 * Copyright (C) 2012, 2013 Giulio Orru`
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package net.sourceforge.andsys;

import android.Manifest.permission;
import android.content.pm.ApplicationInfo;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;
import android.content.pm.PermissionInfo;
import android.graphics.drawable.Drawable;

import java.io.File;
import java.util.Comparator;

/*
 * # Performance Tips:
 *  - Avoid Creating Unnecessary Objects
 *  - Prefer Static Over Virtual
 *  - Use Static Final For Constants
 *  - Avoid Internal Getters/Setters
 *  - Use Enhanced For Loop Syntax
 *  - Consider Package Instead of Private Access with Private Inner Classes
 *  - Avoid Using Floating-Point
 *  - Know and Use the Libraries
 *  - Use Native Methods Carefully
 *  - Performance Myths
 *  - Always Measure
 */

/*
 * TODO Android Package Size (http://www-jo.se/f.pfleger/android-package-size)
 * Unfortunately there is currently no public api to retrieve size information about application 
 * packages. However, building on a technique I wrote about in the past, we can use the 
 * PackageManager's hidden getPackageSize method to retrieve an instance of PackageStats that 
 * includes cache, code and data size information.
 * First of all our application will need the GET_PACKAGE_SIZE permission...
 */

public class App implements Comparable<App> {

    /**
     * Label or Name [USED IN equals METHOD]
     */
    private String mName;

    /**
     * Package Name [USED IN equals METHOD]
     */
    private String mPack;

    /**
     * Version Name (not Number) [USED IN equals METHOD]
     */
    private String mVers;

    /**
     * When the application was installed (Installed Date)
     */
    private long mInst;

    /**
     * Package Icon
     */
    private Drawable mIcon;

    /**
     * Launcher Activity Exists?
     */
    private boolean mFlagLaunchable;
    
    /**
     * Start on boot?
     */
    private boolean mFlagBootable;

    /**
     * Flags of ApplicationInfo
     * 
     * see: FLAG_SYSTEM (Added in API level 1), FLAG_UPDATED_SYSTEM_APP (Added in API level 4),
     * FLAG_EXTERNAL_STORAGE (Added in API level 8), FLAG_INSTALLED (Added in API level 17),
     * FLAG_IS_DATA_ONLY (Added in API level 17).
     */
    private int mFlags;

    /**
     * Permissions Alert
     */
    private int mPermissions;

    /**
     * Constructor
     * 
     * Warning: Constructors should not call an overridable method - that is, they should only call
     * methods that are private, static, or final.
     * 
     * @param pm
     *            PackageManager, for retrieving various kinds of information related to the
     *            application packages that are currently installed on the device.
     * @param pi
     *            PackageInfo
     */
    public App(PackageManager pm, PackageInfo pi) {

        this.mName = pi.applicationInfo.loadLabel(pm).toString();
        this.mPack = pi.packageName;
        this.mVers = pi.versionName;

        // Full path to the location of this package
        {
            String dirPath = pi.applicationInfo.sourceDir;
            File dirResource = new File(dirPath);
            this.mInst = dirResource.lastModified();
        }

        // Icon
        {
            try {
                this.mIcon = pm.getApplicationIcon(this.mPack);
            } catch (NameNotFoundException e) {
                this.mIcon = pm.getDefaultActivityIcon();
            }
        }

        // Launcher Activity Exists?
        this.mFlagLaunchable = (pm.getLaunchIntentForPackage(this.mPack) != null);

        // Flags of ApplicationInfo
        this.mFlags = pi.applicationInfo.flags;

        // Number of Permissions
        this.mPermissions = 0;
        if (pi.requestedPermissions != null) {
            // issue NullPointerException
            this.mPermissions = pi.requestedPermissions.length;

            /*
             * looking for dangerous & start on boot (RECEIVE_BOOT_COMPLETED)
             */
            this.mFlagBootable = false;
            for (String name : pi.requestedPermissions) {
                try {
                    if (pm.getPermissionInfo(name, 0).protectionLevel == PermissionInfo.PROTECTION_DANGEROUS) {
                        // dangerous permission found
                        // number of permission has 4 digits:
                        // n0 n1 n2 n3 ==> n0 & n1 dangerous count; n2 & n3 normal count
                        this.mPermissions = this.mPermissions -1 + 100;

                    } else if (name.equals(permission.RECEIVE_BOOT_COMPLETED)) {
                        this.mFlagBootable = true;
                    }
                } catch (PackageManager.NameNotFoundException e) {
                    // nothing
                }
            }
        }

    }

    public final String getName() {
        return this.mName;
    }

    public final String getPackage() {
        return this.mPack;
    }

    public final String getVersion() {
        return this.mVers;
    }

    public final long getInstalled() {
        return this.mInst;
    }

    public final Drawable getIcon() {
        return mIcon;
    }

    public final boolean getFlagLaunchable() {
        return mFlagLaunchable;
    }
    
    public final boolean getFlagBootable() {
        return mFlagBootable;
    }

    public final boolean getFlagSystem() {
        return ((this.mFlags & ApplicationInfo.FLAG_SYSTEM) == 1);
    }

    public final int getPermissions() {
        return mPermissions;
    }

    /*
     * Compares this instance with the specified object and indicates if they are equal.
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (obj == null || obj.getClass() != this.getClass()) {
            return false;
        }

        App p = (App) obj;
        return (mName == p.mName || (mName != null && mName.equals(p.mName)))
                && (mPack == p.mPack || (mPack != null && mPack.equals(p.mPack)))
                && (mVers == p.mVers || (mVers != null && mVers.equals(p.mVers)));
    }

    /*
     * Returns an integer hash code for this object. By contract, any two objects for which
     * equals(Object) returns true must return the same hash code value.
     * 
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        final int PRIME = 11;
        int result = 2;

        result = PRIME * result + ((mName == null) ? 0 : mName.hashCode());
        result = PRIME * result + ((mPack == null) ? 0 : mPack.hashCode());
        result = PRIME * result + ((mVers == null) ? 0 : mVers.hashCode());

        return result;
    }

    /*
     * Returns a negative integer, zero, or a positive integer as this object is less than, equal
     * to, or greater than the specified object.
     * 
     * Order by: mName, mPackage, mVersion
     * 
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    @Override
    public int compareTo(App another) {
        int compare = 0;

        compare = this.mName.compareTo(another.mName);
        if (compare == 0) {
            compare = this.mPack.compareTo(another.mPack);
            if (compare == 0) {
                compare = this.mVers.compareTo(another.mVers);
            }
        }

        return compare;
    }

    /**
     * Order by "Name". A Comparator is used to compare two objects to determine their ordering with
     * respect to each other.
     */
    protected static class ComparatorName implements Comparator<App> {
        @Override
        public int compare(App lhs, App rhs) {
            // return lhs.mName.compareTo(rhs.mName);
            return lhs.mName.compareToIgnoreCase(rhs.mName);
        }
    }

    /**
     * Order by "Package Name". A Comparator is used to compare two objects to determine their
     * ordering with respect to each other.
     */
    protected static class ComparatorPackage implements Comparator<App> {
        @Override
        public int compare(App lhs, App rhs) {
            // return lhs.mPack.compareTo(rhs.mPack);
            return lhs.mPack.compareTo(rhs.mPack);
        }
    }

    /**
     * Order by "Installed Date". A Comparator is used to compare two objects to determine their
     * ordering with respect to each other.
     */
    protected static class ComparatorInstalled implements Comparator<App> {
        @Override
        public int compare(App lhs, App rhs) {
            int r = 0;
            if (lhs.mInst < rhs.mInst) {
                r = -1;
            } else if (lhs.mInst == rhs.mInst) {
                r = 0;
            } else {
                r = 1;
            }
            return r;
        }
    }

    /**
     * Order by "Installed Date". A Comparator is used to compare two objects to determine their
     * ordering with respect to each other.
     */
    protected static class ComparatorPermissions implements Comparator<App> {
        @Override
        public int compare(App lhs, App rhs) {
            int r = 0;
            if (lhs.mPermissions < rhs.mPermissions) {
                r = -1;
            } else if (lhs.mPermissions == rhs.mPermissions) {
                r = 0;
            } else {
                r = 1;
            }
            return r;
        }
    }

}
